// Generated by CoffeeScript 2.5.1
(function() {
  var CLICKHOUSE_INPUT_FORMAT, CLICKHOUSE_SETTINGS, CLUSTER_WORKER_ID, Cargo, DEFAULT_COMMIT_INTERVAL, DEFAULT_MAX_INSET_PARTS, EXTNAME_UNCOMMITTED, FILENAME_PREFIX, MAX_COMMIT_PER_EXAM_ROUTINE, MIN_ROWS, MIN_TIME, Readable, StaticCountWithinProcess, Transform, assert, cluster, debuglog, fs, fsAsync, isThisLeader, merge2, path, split2;

  fs = require("fs");

  path = require("path");

  cluster = require('cluster');

  assert = require("assert");

  ({isThisLeader} = require("./leader_election"));

  CLUSTER_WORKER_ID = cluster.isMaster ? "nocluster" : cluster.worker.id;

  debuglog = require("debug")(`chcargo:cargo@${CLUSTER_WORKER_ID}`);

  ({Readable, Transform} = require('stream'));

  merge2 = require('merge2');

  split2 = require('split2');

  // Use fs.promises API for async file operations
  fsAsync = fs.promises;

  FILENAME_PREFIX = "cargo_";

  EXTNAME_UNCOMMITTED = ".uncommitted";

  MAX_COMMIT_PER_EXAM_ROUTINE = 1;

  MIN_TIME = 1000;

  MIN_ROWS = 100;

  DEFAULT_COMMIT_INTERVAL = 5000;

  DEFAULT_MAX_INSET_PARTS = 100;

  StaticCountWithinProcess = 0;

  CLICKHOUSE_SETTINGS = {
    wait_end_of_query: 1
  };

  CLICKHOUSE_INPUT_FORMAT = 'JSONCompactEachRow';

  Cargo = class Cargo {
    toString() {
      return `[Cargo ${this.tableName}]`;
    }

    // @param SQLInsertString tableName - Name of the ClickHouse table to insert into
    // @param Object options - Configuration options
    //   @param {String} options.pathToCargoFolder - Path to folder for storing cargo files
    //   @param {Number} [options.maxTime=1000] - Max time in ms before flushing cached rows
    //   @param {Number} [options.maxRows=100] - Max number of rows to cache before flushing
    //   @param {Number} [options.commitInterval=5000] - Interval in ms between commits to ClickHouse
    //   @param {Number} [options.maxInsetParts=100] - Max number of files to combine in one insert
    // @param {Object} [clickhouseClient] - Clickhouse client instance
    constructor(tableName, options = {}, clickhouseClient) {
      this.tableName = tableName;
      this.maxTime = Math.max(parseInt(options.maxTime) || MIN_TIME, MIN_TIME);
      this.maxRows = Math.max(parseInt(options.maxRows) || MIN_ROWS, MIN_ROWS);
      this.commitInterval = Math.max(parseInt(options.commitInterval) || DEFAULT_COMMIT_INTERVAL, this.maxTime);
      this.maxInsetParts = Math.max(parseInt(options.maxInsetParts) || DEFAULT_MAX_INSET_PARTS, 1);
      // Configure paths and HTTP options
      assert(options.pathToCargoFolder, "missing required option: pathToCargoFolder");
      this.pathToCargoFolder = options.pathToCargoFolder;
      this.pathToCargoFile = path.join(this.pathToCargoFolder, FILENAME_PREFIX + this.tableName);
      // TODO: Consider adding retry mechanism for failed inserts
      this.clickhouseClient = clickhouseClient;
      debuglog(`[new Cargo] @tableName:${this.tableName}, @maxTime:${this.maxTime}, @maxRows:${this.maxRows}, @commitInterval:${this.commitInterval}, @pathToCargoFile:${this.pathToCargoFile}`);
      // Initialize state
      this.cachedRows = [];
      this.lastFlushAt = Date.now();
      this.lastCommitAt = Date.now();
      this.countRotation = 0;
      return;
    }

    // push row insert into memory cache
    push() {
      var arr, i, item, j, len;
      arr = Array.from(arguments);
      assert(arr.length > 0, "blank row can not be accepted.");
      for (i = j = 0, len = arr.length; j < len; i = ++j) {
        item = arr[i];
        if (item instanceof Date) {
          arr[i] = Math.round(item.getTime() / 1000);
        } else if (typeof b === "boolean") {
          arr[i] = Number(b);
        }
      }
      this.cachedRows.push(JSON.stringify(arr));
    }

    // flush memory cache to the disk file
    // @param forced Boolean, is force to flush file
    async flushToFile(forced) {
      var err, rowsToFlush;
      // debuglog("#{@} [flushToFile] @_isFlushing:", @_isFlushing)
      if (this._isFlushing) {
        return;
      }
      if (!(this.cachedRows.length > 0)) {
        //debuglog("#{@} [flushToFile] nothing to flush")
        this.lastFlushAt = Date.now();
        return;
      }
      if (!(forced || (this.cachedRows.length > this.maxRows) || (Date.now() > this.lastFlushAt + this.maxTime))) {
        debuglog(`${this} [flushToFile] SKIP threshold not reach`);
        return;
      }
      this._isFlushing = true;
      rowsToFlush = this.cachedRows;
      this.cachedRows = [];
      debuglog(`${this} [flushToFile] -> ${rowsToFlush.length} rows`);
      try {
        await fsAsync.appendFile(this.pathToCargoFile, rowsToFlush.join("\n") + "\n");
      } catch (error) {
        err = error;
        debuglog(`${this} [flushToFile] ${rowsToFlush.length} rows FAILED error:`, err);
        this.cachedRows = rowsToFlush.concat(this.cachedRows); // unshift data back
      }
      debuglog(`${this} [flushToFile] SUCCESS ${rowsToFlush.length} rows`);
      this.lastFlushAt = Date.now();
      this._isFlushing = false;
    }

    // flush memory cache to the disk file
    flushSync() {
      var rowsToFlush;
      if (!(this.cachedRows.length > 0)) {
        debuglog(`${this} [flushSync] nothing to flush`);
        return;
      }
      rowsToFlush = this.cachedRows;
      this.cachedRows = [];
      debuglog(`${this} [flushSync] ${rowsToFlush.length} rows`);
      fs.appendFileSync(this.pathToCargoFile, rowsToFlush.join("\n") + "\n");
    }

    // upload a single cargo file to clickhouse
    async uploadCargoFile(filepath) {
      debuglog(`[uploadCargoFile ${this.tableName}] filepath:`, filepath);
      return (await this.clickhouseClient.insert({
        table: this.tableName,
        format: CLICKHOUSE_INPUT_FORMAT,
        values: fs.createReadStream(filepath),
        clickhouse_settings: CLICKHOUSE_SETTINGS
      }));
    }

    // upload multiple local uncommits to clickhouse
    async uploadLocalUncommitsToClickhouse(filenamList) {
      var combinedStream, filepath, j, len, listOfFileStreams;
      debuglog(`[uploadLocalUncommitsToClickhouse] filenamList ${filenamList.length} :`, filenamList);
      if (!(Array.isArray(filenamList) && filenamList.length > 0)) {
        return;
      }
      listOfFileStreams = filenamList.map(function(filepath) {
        return fs.createReadStream(filepath, {
          encoding: 'utf8'
        }).pipe(split2(JSON.parse));
      });
      combinedStream = merge2(...listOfFileStreams, {
        objectMode: true
      });
      debuglog(`[uploadLocalUncommitsToClickhouse] combinedStream  ${listOfFileStreams.length}`);
      await this.clickhouseClient.insert({
        table: this.tableName,
        format: CLICKHOUSE_INPUT_FORMAT,
        values: combinedStream,
        clickhouse_settings: CLICKHOUSE_SETTINGS
      });
      for (j = 0, len = filenamList.length; j < len; j++) {
        filepath = filenamList[j];
        debuglog("[uploadLocalUncommitsToClickhouse] clean local commit:", path.basename(filepath));
        await fsAsync.unlink(filepath); // remove successfully committed local files
      }
    }

    // check if to commit disk file to clickhouse DB
    async exam() {
      var err, hasRotation;
      try {
        await this.flushToFile();
      } catch (error) {
        err = error;
        debuglog(`[exam ${this.tableName}] ABORT fail to flush. error:`, err);
        return;
      }
      if (!(Date.now() > this.lastCommitAt + this.commitInterval)) {
        return;
      }
      //debuglog "[exam #{@tableName}] SKIP tick not reach"
      if (!isThisLeader()) {
        debuglog(`[exam ${this.tableName}] CANCLE NOT leadWorkerId`);
        // non-leader skip 10 commit rounds
        this.lastCommitAt = Date.now() + this.commitInterval * 10;
        return;
      }
      try {
        //debuglog "[exam] LEAD COMMIT"
        hasRotation = (await this.rotateFile());
        //if hasRotation or @countRotation < 1
        // LAZY: commit only when: 1. has local rotated uncommits, or 2. first time exame to cargo to restore any previous local uncommits
        this.countRotation += Number(hasRotation);
        await this.commitToClickhouseDB();
        this.lastCommitAt = Date.now();
      } catch (error) {
        err = error;
        debuglog(`[exam ${this.tableName}] FAILED to commit. error:`, err);
      }
    }

    // prepar all uncommitted local files
    // @return Boolean, true if there are local uncommits exist
    async rotateFile() {
      var err, pathToRenameFile, stats;
      if (this._isFileRotating) {
        debuglog("[rotateFile] SKIP @_isFileRotating");
        return false;
      }
      this._isFileRotating = true; // lock on
      try {
        stats = (await fsAsync.stat(this.pathToCargoFile));
      } catch (error) {
        err = error;
        if (err.code !== 'ENOENT') {
          debuglog("[rotateFile] ABORT fail to stats file. error:", err);
        }
        //else
        //debuglog "[rotateFile] SKIP nothing to rotate"
        this._isFileRotating = false; // lock released
        return false;
      }
      //debuglog "[rotateFile > stat] err:", err,", stats:", stats
      if (!(stats && (stats.size > 0))) {
        debuglog("[rotateFile] SKIP empty file.");
        this._isFileRotating = false; // lock released
        return false;
      }
      // rotate disk file
      pathToRenameFile = path.join(this.pathToCargoFolder, `${FILENAME_PREFIX}${this.tableName}.${Date.now().toString(36) + `_${++StaticCountWithinProcess}`}.${CLUSTER_WORKER_ID}${EXTNAME_UNCOMMITTED}`);
      debuglog(`[rotateFile] rotate to ${pathToRenameFile}`);
      try {
        await fsAsync.rename(this.pathToCargoFile, pathToRenameFile);
      } catch (error) {
        err = error;
        debuglog(`[rotateFile] ABORT fail to rename file to ${pathToRenameFile}. error:`, err);
      }
      this._isFileRotating = false;
      return true;
    }

    // commit local rotated files to remote ClickHouse DB
    async commitToClickhouseDB() {
      var err, filenamList, rotationPrefix, splitMark;
      try {
        //debuglog "[commitToClickhouseDB]"
        //if @_isCommiting
        //debuglog "[commitToClickhouseDB] SKIP is committing"
        //return

        //@_isCommiting = true  #lock on
        filenamList = (await fsAsync.readdir(this.pathToCargoFolder));
      } catch (error) {
        err = error;
        //debuglog "[commitToClickhouseDB > readdir] err:", err, ", filenamList:", filenamList
        debuglog("[commitToClickhouseDB > ls] FAILED error:", err);
        return;
      }
      //debuglog "[commitToClickhouseDB > readdir] filenamList:", filenamList
      //@_isCommiting = false  # lock release
      if (!(Array.isArray(filenamList) && (filenamList.length > 0))) {
        return;
      }
      // filter out non-commits
      //debuglog "[commitToClickhouseDB] CANCLE empty filenamList"
      //@_isCommiting = false  # lock release
      rotationPrefix = FILENAME_PREFIX + this.tableName + '.';
      filenamList = filenamList.filter(function(item) {
        return item.startsWith(rotationPrefix) && item.endsWith(EXTNAME_UNCOMMITTED);
      });
      if (!(filenamList.length > 0)) {
        return;
      }
      //debuglog "[commitToClickhouseDB > ls] CANCLE empty valid filenamList"
      //@_isCommiting = false  # lock release
      debuglog(`[commitToClickhouseDB] filenamList(${filenamList.length})`);
      
      // sort filenamList from new to old
      splitMark = `${FILENAME_PREFIX}${this.tableName}.`;
      filenamList.sort(function(a, b) {
        var tsA, tsB;
        tsA = parseInt((a.split(splitMark) || [])[1] || 0, 36);
        tsB = parseInt((b.split(splitMark) || [])[1] || 0, 36);
        //debuglog "[commitToClickhouseDB] tsA:#{tsA} tsB:#{tsB}"
        return tsB - tsA;
      });
      //debuglog "[commitToClickhouseDB] AFTER SORT filenamList(#{filenamList.length})" #, filenamList

      // truncate filenamList if too long
      filenamList.length = Math.min(filenamList.length, this.maxInsetParts);
      //debuglog "[commitToClickhouseDB] after truncate filenamList(#{filenamList.length})" #, filenamList

      // padding full file path
      filenamList = filenamList.map((item) => {
        return path.join(this.pathToCargoFolder, item);
      });
      // submit each local uncommit sequentially
      this.uploadLocalUncommitsToClickhouse(filenamList);
    }

  };

  //@_isCommiting = false  # lock release
  module.exports = Cargo;

}).call(this);
